/*******************************************************************************
**	S3 I/O routines
********************************************************************************
**
   Copyright (C) 1995-1997 by Steffen Seeger
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**	This code was derived from the following source of information:
**	[1]	S3 Vision964 Data Book, (c) 1994 S3 Incorporated
*/

#define	IO_OUT16VAL(v, r)	(((v) << 8) | (r))

#define CRTI	CRTIc
#define CRTD	CRTDc
#define INS1r	INS1rc
#define FCTRLw	FCTRLwc

static kgiu8 MISC_in8(struct kgi_display *dpy)
{
	register GET_PRIVATE;

	if (MEM_VGA) {

		register kgiu8 val = mem_in8(MEM_VGA + MISCr);
		DEBUG3("MISC mem -> %.2x", val);
		return val;

	} else {

		register kgiu8 val = io_in8(IO_VGA + MISCr);
		DEBUG3("MISC io -> %.2x", val);
		return val;
	}

}

static void MISC_out8(struct kgi_display *dpy, kgiu8 val)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_out8(val | MISC_COLOR_IO, MEM_VGA + MISCw);
		DEBUG3("MISC mem <- %.2x", val | MISC_COLOR_IO);

	} else {	

		io_out8(val | MISC_COLOR_IO, IO_VGA + MISCw);
		DEBUG3("MISC io <- %.2x", val | MISC_COLOR_IO);
	}
#ifdef VERIFY_WRITE
	MISC_in8(dpy);
#endif
}

kgiu8 CRT_in8(struct kgi_display *dpy, kgiu8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		kgiu8 val;

		mem_out8(reg, MEM_VGA + CRTI);
		val = mem_in8(MEM_VGA + CRTD);
		DEBUG3("CRT mem %.2x -> %.2x", reg, val);
		return val;

	} else {

		kgiu8 val;

		io_out8(reg, IO_VGA + CRTI);
		val = io_in8(IO_VGA + CRTD);
		DEBUG3("CRT io %.2x -> %.2x", reg, val);
		return val;
	}
}

void CRT_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		DEBUG3("CRT mem %.2x <- %.2x", reg, val);
		mem_out8(reg, MEM_VGA + CRTI);
		mem_out8(val, MEM_VGA + CRTD);

	} else {

		DEBUG3("CRT io %.2x <- %.2x", reg, val);
		io_out16(IO_OUT16VAL(val, reg), IO_VGA + CRTI);
	}
#ifdef VERIFY_WRITE
	CRT_in8(dpy, reg);
#endif
}

static inline kgiu8 ATC_in8(struct kgi_display *dpy, kgiu8 reg)
{
	kgiu8 val;

	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		mem_in8(MEM_VGA + INS1r);
		mem_out8(reg, MEM_VGA +  ATCI); /* reset AFF and setup index */
		val = mem_in8(MEM_VGA + ATCDr);
		
		mem_in8(MEM_VGA + INS1r);
		mem_out8( ATCI_ENABLEDISPLAY, MEM_VGA +  ATCI);
		
		DEBUG3("ATC mem %.2x -> %.2x", reg, val);
		return val;

	} else {
		io_in8(IO_VGA + INS1r);
		io_out8(reg, IO_VGA +  ATCI); /* reset AFF and setup index */
		val=io_in8(IO_VGA + ATCDr);
		io_in8(IO_VGA + INS1r);
		io_out8( ATCI_ENABLEDISPLAY, IO_VGA + ATCI);
		
		DEBUG3("ATC io %.2x -> %.2x", reg, val);
		return val;
	}
}

static void ATC_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_in8(MEM_VGA + INS1r);	/* reset AFF	*/
		mem_out8(reg, MEM_VGA + ATCI);	/* set index	*/
		mem_out8(val, MEM_VGA + ATCDw);	/* write value	*/
		mem_in8(MEM_VGA + INS1r);	/* reset AFF	*/
		mem_out8(ATCI_ENABLEDISPLAY, MEM_VGA + ATCI);
		DEBUG3("ATC mem %.2x <- %.2x", reg, val);

	} else {

		io_in8(IO_VGA + INS1r);		/* reset AFF	*/
		io_out8(reg, IO_VGA + ATCI);	/* set index	*/
		io_out8(val, IO_VGA + ATCDw);	/* write value	*/
		io_in8(IO_VGA + INS1r);		/* reset AFF	*/
		io_out8(ATCI_ENABLEDISPLAY, IO_VGA + ATCI);
		DEBUG3("ATC io %.2x <- %.2x", reg, val);
	}
#ifdef VERIFY_WRITE
	ATC_in8(dpy, reg);
#endif
}

kgiu8 SEQ_in8(struct kgi_display *dpy, kgiu8 reg)
{
	GET_PRIVATE;
	
	if (MEM_VGA) {

		kgiu8 val;

		mem_out8(reg, MEM_VGA + SEQI);
		val = mem_in8(MEM_VGA + SEQD);
		DEBUG3("SEQ mem %.2x -> %.2x", reg, val);
		return val;


	} else {
		kgiu8 val;

		io_out8(reg, IO_VGA + SEQI);
		val = io_in8(IO_VGA + SEQD);
		DEBUG3("SEQ io %.2x -> %.2x", reg, val);
		return val;
	}
}

void SEQ_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + SEQI);
		mem_out8(val, MEM_VGA + SEQD);
		DEBUG3("SEQ mem %.2x <- %.2x", reg, val);

	} else {

		io_out16(IO_OUT16VAL(val, reg), IO_VGA + SEQI);
		DEBUG3("SEQ io %.2x <- %.2x", reg, val);
	}
	
#ifdef VERIFY_WRITE
	SEQ_in8(dpy, reg);
#endif
}

static inline kgiu16 GRC_in8(struct kgi_display *dpy, kgiu8 reg)
{
	GET_PRIVATE;

	if (MEM_VGA) {
		kgiu8 val;

		mem_out8(reg, MEM_VGA + GRCI);
		val = mem_in8(MEM_VGA + GRCD);
		DEBUG3("GRC mem %.2x -> %.2x", reg, val);
		return val;

	} else {
		kgiu8 val;
		
		io_out8(reg, IO_VGA + GRCI);
		val = io_in8(IO_VGA + GRCD);
		DEBUG3("GRC io %.2x -> %.2x", reg, val);
		return val;
	}
}

static inline void GRC_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;
	
	if (MEM_VGA) {

		mem_out8(reg, MEM_VGA + GRCI);
		mem_out8(val, MEM_VGA + GRCD);
		DEBUG3("GRC mem %.2x <- %.2x", reg, val);

	} else {

		io_out16(IO_OUT16VAL(val, reg), IO_VGA + GRCI);
		DEBUG3("GRC io %.2x <- %.2x", reg, val);
	}
#ifdef VERIFY_WRITE
	GRC_in8(dpy,reg);
#endif
}

void SPR_out32(struct kgi_display *dpy, kgiu32 offs, kgiu32 val)
{
	GET_PRIVATE;

	if (MEM_VGA) {
		mem_out32(val,MEM_VGA +offs);
		DEBUG3("SPR mem %.4x <- %.2x", offs,val);
	}
	else
	  DEBUG0("Error ! MEM_VGA not valid.");
}

static void s3_unlock(struct kgi_display *dpy)
{
	register GET_PRIVATE;

	if (MEM_VGA) {

		#if defined(S3_86C968) || defined(S3_86C868)

			/*	MMIO only wake-up
			*/
			mem_out8(0x10, MEM_LAW + 0x1008508);
			mem_out8(0x01, MEM_LAW + 0x1008502);
			mem_out8(0x08, MEM_LAW + 0x1008508);

			NOTICE2("wakeup");

		#elif defined(S3_86C765) || defined(S3_86C325)

			mem_out8(0x01, MEM_LAW + 0x10083c3);

		#elif !(defined(S3_86C928))

			#warning check the code above!
		#endif

	} else {

		#if defined(S3_86C964) || defined(S3_86C968) || \
		    defined(S3_86C864) || defined(S3_86C868) || \
		    defined(S3_86C928)

			/*	928 has I/O only, but may need explicit wakeup.
			*/
			#warning 0x102 not claimed!
			io_out8(0x10, IO_VGA + S3_SUBSYSTEM);
			io_out8(0x01, IO_VGA + S3_SETUP);
			io_out8(0x08, IO_VGA + S3_SUBSYSTEM);

		#elif defined(S3_86C765) || defined(S3_86C325)

			io_out8(0x01, IO_VGA + 0x3C3);
 		#endif
	}

	MISC_out8(dpy, MISC_in8(dpy) | MISC_COLOR_IO);

	CRT_out8(dpy, CRT_in8(dpy, 0x11) & 0x7f, 0x11);
	CRT_out8(dpy, 0x48, 0x38);		/* unlock CR30-CR3F */
	CRT_out8(dpy, 0xA5, 0x39);		/* unlock CR40-CRFF */
	CRT_out8(dpy, CRT_in8(dpy, 0x33) & ~(CR33_UNLOCK_VDE | CR33_LOCK_DAC |
		CR33_LOCK_PALETTE), 0x33);

	#if defined(S3_86C325)
 	  /* The XFree server doesn't touch this, though it must be set
	     in the docs. It crashes my system, so S3 is wrong here :) */
	  /* CRT_out8(dpy, 0, 0x35); */
	  /* XFree says S3 again inverted this bits below in the docs...
	     I believe XFree, cause XFree does work, and I never saw a book
	     driving my videocard... */ 
	  CRT_out8(dpy, CRT_in8(dpy, 0x40) & ~0x01, 0x40);	
 	#endif

	SEQ_out8(dpy, SR08_MAGIC, 0x08);

	/*	there are more locks, but this should work after reboot.
	*/
}

static void s3_lock(struct kgi_display *dpy)
{
	CRT_out8(dpy, CRT_in8(dpy, 0x11) | 0x01, 0x11);
	CRT_out8(dpy, 0, 0x38);			/* lock CR30-CR3F */
	CRT_out8(dpy, 0, 0x39);			/* lock CR40-CRFF */
}

/*
**	generalized DAC output functions
*/

#if defined(S3_86C928)

	#define	VRAM_DAC
	#undef	BUGGY_MMIO
#endif

#if defined(S3_86C968) || defined(S3_86C964) || \
    defined(S3_86C864) || defined(S3_86C868)

	#define	VRAM_DAC
	#define	BUGGY_MMIO
#endif

#if defined(S3_86C765) || defined(S3_86C325)

	#define	INTERNAL_DAC
#endif

static kgiu16 dac_io[4]  = { 0x3C8, 0x3C9, 0x3C6, 0x3C7 };
#ifdef	BUGGY_MMIO

	static kgiu16 dac_mem[4] = { 0x3C8, 0x3C9, 0x3CA, 0x3CB };
#else
	static kgiu16 dac_mem[4] = { 0x3C8, 0x3C9, 0x3C6, 0x3C7 };
#endif

#ifdef VRAM_DAC

	#define	DAC_REGISTERS	0x10	/* RS[0:3] are possible	*/
#endif

#ifdef INTERNAL_DAC

	#define	DAC_REGISTERS	0x08	/* RS[0:2] are possible */
#endif


void DAC_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif
		mem_out8(val, MEM_VGA + dac_mem[reg & 0x03]);

//		DEBUG3("DAC mem %.2x <- %.2x", reg, val);

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_out8(val, IO_VGA + dac_io[reg & 0x03]);

//		DEBUG3("DAC io %.2x <- %.2x", reg, val);
	}
}

kgiu8 DAC_in8(struct kgi_display *dpy, kgiu8 reg)
{
	#if !defined(S3_86C928)
		GET_PRIVATE;
	#endif
	register kgiu8 foo;
  
	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
//			DEBUG2("CR52 = %.2x,", mem_in8(MEM_VGA + CRTD));
		#endif

		foo = mem_in8(MEM_VGA + dac_mem[reg & 0x03]);
//		DEBUG3("DAC mem %.2x -> %.2x", reg, foo);
		return foo;

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, CRTI);
			io_out8((io_in8(CRTD) & ~CR55_DAC_MASK) | (reg >> 2),
				CRTD);
		#endif
	
		foo = io_in8(dac_io[reg & 0x03]);
//		DEBUG3("DAC io %.2x -> %.2x", reg, foo);
		return foo;
	}
}

void DAC_ins8(struct kgi_display *dpy, kgiu8 reg, void *buf, kgi_uint count)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif

		mem_ins8(MEM_VGA + dac_mem[reg & 0x03], buf, count);

	} else {


		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_ins8(IO_VGA + dac_mem[reg & 0x03], buf, count);
	}
}

void DAC_outs8(struct kgi_display *dpy, kgiu8 reg, void *buf, kgi_uint count)
{
	GET_PRIVATE;

	ASSERT(reg < DAC_REGISTERS);

	if (MEM_VGA) {

		#if DAC_REGISTERS > 8

			mem_out8(0x55, MEM_VGA + CRTI);
			mem_out8((mem_in8(MEM_VGA+CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), MEM_VGA + CRTD);
		#endif

		mem_outs8(MEM_VGA + dac_mem[reg & 0x03], buf, count);

	} else {

		#if DAC_REGISTERS > 8

			io_out8(0x55, IO_VGA + CRTI);
			io_out8((io_in8(IO_VGA + CRTD) & ~CR55_DAC_MASK) |
				(reg >> 2), IO_VGA + CRTD);
		#endif

		io_outs8(IO_VGA + dac_io[reg & 0x03], buf, count);
	}
}

/*
**	clock I/O functions
*/

#if defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C928) \
 || defined(S3_86C864) || defined(S3_86C868)


#define	CLK_DECODE	(CR42_DCLK_MASK+1)

void CLK_out8(struct kgi_display *dpy, kgiu8 value, kgiu8 reg)
{
	ASSERT(value < CLK_DECODE);

	CRT_out8(dpy, (CRT_in8(dpy, 0x42) & ~CR42_DCLK_MASK) | value, 0x42);
} 

kgiu8 CLK_in8(struct kgi_display *dpy, kgiu8 reg)
{
	return CRT_in8(dpy, 0x5C) & CR5C_CLOCK_MASK;
}

#endif

#if defined(S3_86C765) || defined(S3_86C325)

void CLK_out8(struct kgi_display *dpy, kgiu8 val, kgiu8 reg)
{
	if (reg < 7) {

		switch (reg) {

		case 0:
			val = (val & 3) << 2;
			MISC_out8(dpy, (MISC_in8(dpy) & ~MISC_CLOCK_MASK) |
				val);
			return;

		case 6:
		  	SEQ_out8(dpy, (SEQ_in8(dpy, 0x15) & ~CLK6_CLOCK_MASK) |
				(val & CLK6_CLOCK_MASK), 0x15);
			return;

		default:
			SEQ_out8(dpy, val, reg + 0x0F);
		}

	} else {

		DEBUG0("invalid CLK register index (%i)", reg);
	}
}

kgiu8 CLK_in8(struct kgi_display *dpy, kgiu8 reg)
{
	if (reg < 7) {

		switch (reg) {

		case 0:
			return (MISC_in8(dpy) >> 2) & 3;

		case 6:
			return SEQ_in8(dpy, 0x15) & ~CLK6_CLOCK_MASK;

		default:
			return SEQ_in8(dpy, reg + 0x0F);
		}

	} else {

		DEBUG0("invaluid CLK register index (%i)", reg);
		return 0xFF;
	}
}

#endif

/*
**	accelerator control
*/

#ifndef ViRGE
 
static inline void AC_out16(struct kgi_display *dpy, kgiu16 val, kgiu16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW points to the new MMIO region.
	*/
	if (MEM_VGA) {

		switch (reg) {

		case 0x42e8:
			mem_out16(val, MEM_LAW + 0x1008504);
			return;

		case 0x46e8:
			mem_out16(val, MEM_LAW + 0x1008508);
			return;

		case 0x4ae8:
			mem_out16(val, MEM_LAW + 0x100850c);
			return;
		}

	} else {

		io_out16(val, IO_ACCEL + reg);
	}
}

static inline kgiu16 AC_in16(struct kgi_display *dpy, kgiu16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW will point to a new-MMIO region.
	*/
	if (MEM_VGA) {

		switch (reg) {

		case AC_SUBSYS_STAT:
			return mem_in16(MEM_LAW + 0x1008504);
		case 0x46e8:
			return mem_in16(MEM_LAW + 0x1008508);
		case AC_ADVFUNC_CNTL:
			return mem_in16(MEM_LAW + 0x100850c);
		}
		return 0xFFFF;

	} else {

		return io_in16(IO_ACCEL + reg);
	}
}
#endif
/*
**	accelerator setup
*/

static inline void ACCEL_out16(struct kgi_display *dpy, kgiu16 val, kgiu16 reg)
{
	GET_PRIVATE;

	if (MEM_ACCEL) {

		mem_out16(val, MEM_ACCEL + reg);

	} else {

		io_out16(val, IO_ACCEL + reg);
	}
}


static inline kgiu16 ACCEL_in16(struct kgi_display *dpy, kgiu16 reg)
{
	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		ASSERT(MEM_ACCEL);
		return mem_in16(MEM_ACCEL + reg);

	} else {

		return io_in16(IO_ACCEL + reg);
	}
}


/*
**	State setting routines.
*/

struct statetable
{
	kgiu8	MISC;
	kgiu8	FCTRL;
	kgiu8	SEQ[NoSEQRegs];
	kgiu8	CRT[NoCRTRegs];
	kgiu8	GRC[NoGRCRegs];
	kgiu8	ATC[NoATCRegs];
};

static void save_ATC(struct kgi_display *dpy, kgiu8 *buf)
{
	kgiu8 i;

	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		for(i = 0; i < NoATCRegs; i++) { 
			mem_in8(MEM_VGA + INS1r); mem_out8(i,MEM_VGA +  ATCI); /* reset AFF and setup index */
			*(buf++)=mem_in8(MEM_VGA + ATCDr);
		}
		mem_in8(MEM_VGA + INS1r); mem_out8( ATCI_ENABLEDISPLAY,MEM_VGA +  ATCI);

	} else {
		for(i = 0; i < NoATCRegs; i++) { 
			io_in8(IO_VGA + INS1r); io_out8(i,IO_VGA +  ATCI); /* reset AFF and setup index */
			*(buf++)=io_in8(IO_VGA + ATCDr);
		}
		io_in8(IO_VGA + INS1r); io_out8( ATCI_ENABLEDISPLAY, IO_VGA + ATCI);

	}
}

static void save_GRC(struct kgi_display *dpy, kgiu8 *buf)
{
	kgiu8 i;

	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		for(i = 0; i < NoGRCRegs; i++) { 
			mem_out8(i,MEM_VGA +  GRCI);
			*(buf++)=mem_in8(MEM_VGA + GRCD);
		}

	} else {
		for(i = 0; i < NoGRCRegs; i++) { 
			io_out8(i, IO_VGA + GRCI);
			*(buf++)=io_in8(IO_VGA + GRCD);
		}

	}
}

static void save_CRT(struct kgi_display *dpy, kgiu8 *buf)
{

	kgiu8 i;

	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/

	if (MEM_VGA) {

		for(i = 0; i < NoCRTRegs; i++) { 
			mem_out8(i,MEM_VGA +  CRTI);
			*(buf++)=mem_in8(MEM_VGA + CRTD);
		}

	} else {
		for(i = 0; i < NoCRTRegs; i++) { 
			io_out8(i, IO_VGA + CRTI);
			*(buf++)=io_in8(IO_VGA + CRTD);
		}

	}
}

static void save_SEQ(struct kgi_display *dpy, kgiu8 *buf)
{
	kgiu8 i;

	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		for(i = 0; i < NoSEQRegs; i++) { 
			mem_out8(i, MEM_VGA + SEQI);
			*(buf++)=mem_in8(MEM_VGA + SEQD);
		}

	} else {
		for(i = 0; i < NoSEQRegs; i++) { 
			io_out8(i, IO_VGA + SEQI);
			*(buf++)=io_in8(IO_VGA + SEQD);
		}

	}
}

static void save_state(struct kgi_display *dpy, struct statetable *state)
{
	GET_PRIVATE;

	//s3_unlock(dpy);

	state->MISC = MISC_in8(dpy);

	if (MEM_VGA) {
		state->FCTRL = mem_in8(MEM_VGA + FCTRLr);/* save feature control reg */
	} else {
		state->FCTRL = io_in8(IO_VGA + FCTRLr);	/* save feature control reg */
	}
	save_SEQ(dpy,state->SEQ);	/* save sequencer regs */
	save_CRT(dpy,state->CRT);	/* save cathode ray tube controller regs */
	save_GRC(dpy,state->GRC);	/* save graphics controller regs */
	save_ATC(dpy,state->ATC);	/* save attribute controller regs */
}

#define SLOW_RESTORE
static void restore_SEQ(struct kgi_display *dpy, kgiu8 *buf)
{
	register kgiu8 i;

#ifndef SLOW_RESTORE
	if (MEM_VGA) {

		for (i = 0; i < NoSEQRegs; i++) {

			mem_out8(i, MEM_VGA + SEQI);
			mem_out8(*(buf++), MEM_VGA + SEQD);
		}

	} else {
	
		for (i = 0 ; i < NoSEQRegs; i++) {

			io_out16(IO_OUT16VAL(*(buf++), i), IO_VGA + SEQI);
		}
	}
#else
	for (i = 0; i < NoSEQRegs; i++) {
		switch (i) {
		case 5:
		case 6:
		case 7:
		case 8:
		case 9:
		case 0xc:
		case 0xd:
		case 0xe:
		case 0xf:
#ifdef ViRGE
		case 0x10:
		case 0x11:
#endif
		case 0x12:
		case 0x13:
		case 0x15:
		case 0x18:
		case 0x19:
			break;
		default:
			SEQ_out8(dpy, buf[i] ,i);
		}
	}
#endif
}

static void restore_CRT(struct kgi_display *dpy, kgiu8 *buf)
{
	register kgiu8 i;

	CRT_out8(dpy, CRT_in8(dpy, CRT_VSYNCEND) & ~CR11_LOCKTIMING,
		CRT_VSYNCEND);

	for(i = 0x00; i < 0x19; i++) {

		CRT_out8(dpy, buf[i], i);
	}

	for(i = 0x22; i < 0x3D; i++) {
	#ifdef ViRGE
 	  switch (i) {
	    case 0x22: // Read only
	    case 0x23: // For GX / DX only
	    case 0x24: // Read only
	    case 0x25: // Undocumented
	    case 0x26: // For GX / DX only, readonly for other chipsets
	    case 0x27: // Undocumented
	    case 0x28: // Undocumented
	    case 0x29: // Undocumented
	    case 0x2A: // Undocumented
	    case 0x2B: // Undocumented
	    case 0x2C: // Undocumented
	    case 0x2D: // Read only
 	    case 0x2E: // Read only
  	    case 0x2F: // Read only
	    case 0x30: // Read only
	    case 0x32: // Setup by Interrupt handler
            case 0x33: // Setup by unlock
	    case 0x34: // Setup at initialisation
	    case 0x35: // Should be set by startup, but crashes ? 
	    case 0x36: // Don't touch register ?
	    case 0x37: // Setup at initialisation
 	    case 0x38: // Setup by unlock
	    case 0x39: // Setup by unlock
 	      break;
	    default:
	      CRT_out8(dpy, buf[i], i);
	  }
	#endif
	#ifndef ViRGE
		CRT_out8(dpy, buf[i], i);
	#endif
	} 

	for(i = 0x40; i < 0x70; i++) {
 
		switch(i) {

		#if defined(S3_86C964) || defined(S3_86C968) || \
			defined(S3_86C928)

		case 0x42:
			CRT_out8(dpy, (CRT_in8(dpy, i) & CR42_DCLK_MASK) |
				(buf[i] & ~CR42_DCLK_MASK), i); 
			break;

		case 0x55:
			CRT_out8(dpy, (CRT_in8(dpy, i) & CR55_DAC_MASK) |
			  	(buf[i] & ~CR55_DAC_MASK), i);
			break;
		case 0x4a:
		case 0x4b:
		#endif

		#if defined(S3_86C968)
		case 0x59:
		case 0x5A:
			/*	The S3 Vision 86c968 has some trouble when 
			**	accessing these registers in new MMIO mode
			**	(clears the LAW address to 0 when writing CRTD).
			**	This 'locks' the first 64 MB of memory space.
			**	Thus we better avoid it ...
			*/
			{
 				register GET_PRIVATE;

				if (MEM_VGA == NULL) {

					CRT_out8(dpy, buf[i], i);
				}
			}
		case 0x4a:
		case 0x4b:
			break;
		#endif
		#ifdef ViRGE
		case 0x40: // Setup by wakeup
		case 0x41: // Not used
		case 0x44: // Undocumented
		 
		case 0x4A: // ????
		case 0x4B: // ????

		case 0x50: // Needed for ViRGE VX ???
		case 0x52: // Seems obsolete
		case 0x53: // Once set at initialisation
		case 0x54: // Never set yet.
		case 0x55: // Once set at initialisation
		case 0x56: // Once set at initialisation
		case 0x57: // Undocumented
		case 0x58: // Once set at initialisation
		case 0x59: // Setup by PCI BIOS 
		case 0x5A: // Setup by PCI BIOS
		case 0x5B: // Undocumented
		case 0x5C: // Never used
		case 0x5F: // Undocumented
		case 0x60: // Undocumented
		case 0x61: // Needs to be set for big endian
		case 0x62: // Undocumented
		case 0x64: // Undocumented
		case 0x65: // Once set at initialisation
		case 0x67: // Set by ramdac driver
		case 0x68: // Set by power-on. Don't touch
		case 0x69: // Set by set_origin
		case 0x6A: // Obsolete for new MMIO mode
		case 0x6B: // Not used
		case 0x6C: // Not used
		case 0x6D: // Not used
		case 0x6E: // Not used
		case 0x6F: // Set by power-on 
			break;
		#endif
		default: 
			CRT_out8(dpy, buf[i], i);
		}
	}
	#if 0
	for(i = 0x80; i < NoCRTRegs; i++) {
 
		switch(i) {
		default:
			CRT_out8(dpy, buf[i], i);
		}
	}
	#endif

	
}

static void restore_GRC(struct kgi_display *dpy, kgiu8 *buf)
{
	register kgiu8 i;

#ifndef SLOW_RESTORE
	if (MEM_VGA) {

		for (i = 0; i < NoGRCRegs; i++) {

			mem_out8(i, MEM_VGA + GRCI);
			mem_out8(*(buf++), MEM_VGA + GRCD);
		}

	} else {

		for (i = 0; i < NoGRCRegs; i++) {

			io_out16((*(buf++) << 8) | i, IO_VGA + GRCI);
		}
	}
#else
	for (i = 0; i < NoGRCRegs; i++) {
		GRC_out8(dpy, buf[i] ,i);
	}
#endif
}

static void restore_ATC(struct kgi_display *dpy, kgiu8 *buf)
{
	register kgiu8 i;
	GET_PRIVATE;
#ifndef SLOW_RESTORE
	if (MEM_VGA) {

		for (i = 0; i < NoATCRegs; i++) {

			mem_in8(MEM_VGA + INS1r);
			mem_out8(i, MEM_VGA + ATCI);
			mem_out8(*(buf++), MEM_VGA + ATCDw);
		}

		mem_in8(MEM_VGA + INS1r);
		mem_out8(ATCI_ENABLEDISPLAY, MEM_VGA + ATCI);

	} else {

		for (i = 0; i < NoATCRegs; i++) { 

			io_in8(IO_VGA + INS1r);
			io_out8(i, IO_VGA + ATCI);
			io_out8(*(buf++), IO_VGA + ATCDw);
		}

		io_in8(IO_VGA + INS1r);
		io_out8(ATCI_ENABLEDISPLAY, IO_VGA + ATCI);
	}
#else
	for (i = 0; i < NoATCRegs; i++) {
		ATC_out8(dpy, buf[i] ,i);
	}
#endif
}

static void restore_state(struct kgi_display *dpy, struct statetable *state)
{
	GET_PRIVATE;

	MISC_out8(dpy, state->MISC);
	if (MEM_VGA) {
		mem_out8(state->FCTRL, MEM_VGA + FCTRLw);
	} else {
		io_out8(state->FCTRL, IO_VGA + FCTRLw);
	}

	NOTICE2("MISC & FCTRL set");

	restore_SEQ(dpy, state->SEQ);
	NOTICE2("sequencer set");
	restore_CRT(dpy, state->CRT);
	NOTICE2("CRT set");
	restore_GRC(dpy, state->GRC);
	NOTICE2("GRC set");
	restore_ATC(dpy, state->ATC);
	NOTICE2("ATC set");
}
#ifndef ViRGE
static unsigned short fifo_masks[13] =
{
	0x0080, 0x0040, 0x0020, 0x0010, 
	0x0008, 0x0004, 0x0002, 0x0001,
	0x8000, 0x4000, 0x2000, 0x1000,
	0x0800
/*	These are somehow broken for the Vision964

	0x0800, 0x1000, 0x2000, 0x4000, 
	0x8000, 0x0001, 0x0002, 0x0004, 
	0x0008, 0x0010, 0x0020, 0x0040 
*/
};

static inline void wait_fifo(struct kgi_display *dpy, kgi_uint i)
{
	GET_PRIVATE;
	register kgiu16 mask = fifo_masks[(i < 13) ? i : 12];

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {

		ASSERT(MEM_ACCEL);
		while (mask & mem_in16(MEM_ACCEL + ECR_GP_STATUS));

	} else {

		while (mask & io_in16(IO_ACCEL + ECR_GP_STATUS));
	}
}
#endif
static inline void wait_display_off(struct kgi_display *dpy)
{
	GET_PRIVATE;

	if (MEM_VGA) {

		while (~mem_in8(MEM_VGA + INS1r) & INS1_DISP_OFF);

	} else {

		while (~io_in8(IO_VGA + INS1r) & INS1_DISP_OFF);
	}
}

static void dump_state(struct kgi_display *dpy, struct statetable *state)
{
	int i;

	printk("MISC:  %.2x",state->MISC);
	printk("\nFCTRL: %.2x",state->FCTRL);

	printk("\n--- 00  01  02  03  04  05  06  07  08  09  0a  0b  0c  0d  0e  0f");
	printk("\nSEQ:");

	for (i = 0; i < NoSEQRegs; i++) {

		if(! (i & 0xF)) {

			printk("\n%.2x: ", i); 
		}
		printk("%.2x, ", state->SEQ[i]);
	}
	
	printk("\nCRT:");

	for (i = 0; i < NoCRTRegs; i++) {

		if (! (i & 0xf)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->CRT[i]);
	}
	
	printk("\nGRC:");

	for (i = 0; i < NoGRCRegs; i++) {

		if (! (i & 0xF)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->GRC[i]);
	}
	
	printk("\nATC:");

	for (i = 0; i < NoATCRegs; i++) {

		if (! (i & 0xF)) {

			printk("\n%.2x: ",i);
		}
		printk("%.2x, ", state->ATC[i]);
	}
	printk("\n");
}

